(define-extension form-post)
(declare
  (usual-integrations)
  (export
    request->form-field-list
    request->form-field-alist
    form-field-list->form-field-alist
    form-field?
    form-field:name
    form-field:body
    form-field:content-type
    form-field:file-name
    form-field:file-field?))

(require-extension http-server mime)

(define-record-type
  :form-field
  (*make-form-field name body file-name content-type)
  form-field?
  (name form-field:name)
  (body form-field:body)
  (content-type form-field:content-type)
  (file-name form-field:file-name))

(define (make-form-field name body #!optional file-name content-type)
  (*make-form-field name body file-name content-type))

(define (form-field:file-field? form-field)
  (form-field:file-name form-field))

; return a port that wraps src-port and returns EOF when the specified number
; of bytes have been read.  chicken's MIME egg doesn't seem to stop reading
; after the end-marker boundary

(define (make-fixed-length-port-wrapper src-port max-bytes)
  (define bytes-read 0)

  (define (wrapped-read)
    (let ((c (read-char src-port)))
      (set! bytes-read (add1 bytes-read))
      (if (= max-bytes bytes-read)
        #!eof
        c)))

  (make-input-port wrapped-read
    (lambda () (char-ready? src-port))
    (lambda () (close-input-port src-port))))

; turn a set of headers from the alist given to us by the HTTP package into
; something the MIME package will understand, like the rfc8222 package does.
; headers = the headers alist, wanted = a list of string header names
; identifying the headers we want to convert

(define (http-package-headers->rfc822-package-list headers wanted)
  (map (lambda (h)
         (list (string-downcase h)
               (alist-ref-string h headers)))
       wanted))

(define (parse-multipart/form-data content-length headers port)
  (let* ((headers (http-package-headers->rfc822-package-list
                    headers '("content-type"
                              "content-transfer-encoding")))
          (wrapped-port (make-fixed-length-port-wrapper port content-length))
          (parsed-body (mime-parse-message
                        wrapped-port headers mime-body->string)))
    parsed-body))

(http:content-parser 'multipart/form-data
  (lambda (size headers port)
    (values (parse-multipart/form-data size headers port) "")))

(define (alist-ref-string k alist)
  (alist-ref k alist string=?))

(define (mime-part->form-field mime-part)
  (receive
    (dispos params)
    (extract-header-value-and-params (rfc822-header-ref
                                       (mime-part:headers mime-part)
                                       "content-disposition"))
    (if (string=? dispos "form-data")
      (let ((header-ref (lambda (h) (alist-ref-string h params))))
        (make-form-field (header-ref "name")
                         (mime-part:content mime-part)
                         (header-ref "filename")
                         (and-let* ((headers (mime-part:headers mime-part))
                                    (c-type  (alist-ref-string "content-type" headers)))
                           (car c-type))))
      (error
        'mime-part->form-field
        "expected content-disposition of 'form-data'"))))

(define (top-level-mime-part->form-field-list part)
  (map mime-part->form-field (mime-part:content part)))

(define (request->form-field-list req)
  (top-level-mime-part->form-field-list (http:request-body req)))

(define (form-field-list->form-field-alist form-field-list)
  (fold (lambda (field acc)
          (cons (cons (form-field:name field) field) acc))
        '() form-field-list))

(define (request->form-field-alist req)
  (form-field-list->form-field-alist (request->form-field-list req)))

(define (extract-header-value-and-params header)
  (call-with-input-string header
    (lambda (input)
      (values (rfc822-next-token input (list token-chars))
              (get-attributes input '())))))

; token-chars and (get-attributes) taken from mime egg

(define token-chars
  (char-set-difference
    (ucs-range->char-set #x21 #x7f)
    (char-set #\( #\) #\< #\> #\@ #\, #\;
              #\: #\\ #\" #\/ #\[ #\] #\? #\=)))

(define (get-attributes input r)
    (cond ((and-let* (((eqv? #\; (rfc822-next-token input '())))
                      (attr (rfc822-next-token input `(,token-chars)))
                      ((string? attr))
                      ((eqv? #\= (rfc822-next-token input '())))
                      (val  (rfc822-next-token
                              input
                              `(,token-chars
                                  (,(char-set #\") . ,rfc822-quoted-string))))
                      ((string? val)))
                      (cons attr val))
            => (lambda (p) (get-attributes input (cons p r))))
          (else (reverse! r))))
